/*Class to evaluate expression groups, made for DiffEqWeb*/
/*Author: Mika Spra, 2001 @Helsinki University of Technology, Finland*/
/*In MatTa-project: http://www.math.hut.fi/matta*/
/*Commercial use prohibited*/

public class Solver{

	boolean error=false;						/*Error flag*/
	String error_msg="";						/*Error messages*/
   	double[][] numbers_orig = new double[1][1];	/*Data structure for chars*/
	int[][] chars_orig = new int[1][1];			/*Data structure for numbers*/
	double[][] numbers = new double[1][1];		/*Data structure for evaluation*/
	int[][] chars = new int[1][1];				/*Data structure for evaluation*/
	int max_comp = 12;							/*Maximum number of components*/
	int components = 1;							/*Number of components in data structure*/
	String exp_orig = "";						/*Original expression group in*/

/*Inits the data structure and splits the expression group*/
/*into single expressions to be handled by readExp.*/
/*Expressions must be separated by ";".*/
/*Returns false if there was a problem and true in all went fine*/

	public final boolean parser(String expression){
		int i;		
		boolean status = false;
		int begin = 0;
		int comp = 1;
		if (expression.length() > 0){
			if ((expression.charAt(expression.length()-1)) != ';'){
				for (i = 0; i<expression.length(); i++){
					if (expression.charAt(i)==';')
						comp = comp + 1;
				}
				if ((comp < max_comp+1) && (comp > 0)){
					status = true;
					int exp_length=expression.length() + 5;
					numbers_orig = new double[exp_length][comp];
					chars_orig = new int[exp_length][comp];
					numbers = new double[exp_length][comp];
					chars = new int[exp_length][comp];
					/*Sets the component count*/
					components = comp;
					comp = 0;
					for (i = 0; i<expression.length(); i++){
						if (expression.charAt(i)==';'){
							status = readExp(expression.substring(begin,i),comp);
							begin = i+1;
							comp++;
							if (status == false) {
								error_msg="Bad input";
							 	break;
							}
						}
					}
					if (status)
						status = readExp(expression.substring(begin,i),comp);
				}
			}
		}
		if (status == false) {
			error = true;
			if (exp_orig.length()>1) 
				parser(exp_orig);
		}
		else 
			exp_orig=expression;
		return status;
	}

/*Reads single expressions into data structure*/
/*Data structure consists of 2 tables, one for numbers and one for*/
/*Functions, operators and variables. The latter is coded into integers:*/
/*'^'=1,'*'=2,'/'=3,'+'=4,'-'=5,'='=6,'('=7,')'=8,'sin'=9,'cos'=10*/
/*'tan'=11,'y'=12,'t'=13,'log'=14,'exp'=15,'abs'=16,'sqrt'=17,
/*'asin'=18,'acos'=19,'atan'=20,'sign'=21, number=99.*/
/*Possible variables are t,y=y0,y1,y2,y3,y4,y5,y6,y7,y8,y9,y10,y11*/
/*depending on the number of expressions.*/

	public final boolean readExp(String exp0, int comp) {
		boolean status = true;
		String operators="^*/+-=()";
		int begin_num = 0;
		int i = 1;
		int oper_i= 0;
		int left_brac = 1;
		int right_brac = 0;
		boolean fun=false;
		String number;
		String expression = exp0 + ")=";
		chars_orig[0][comp] = 7;
		numbers_orig[0][comp] = 0;
		if (expression.charAt(oper_i) == '-') {
			chars_orig[1][comp] = 99;
			numbers_orig[1][comp] = 0;
			i = 2;
		}
		while(oper_i<expression.length() && (status == true)) {
			if (operators.indexOf(expression.charAt(oper_i)) ==-1)
				oper_i++;
			else {
				number=expression.substring(begin_num,oper_i).trim();
				if(number.length()>0) {
					numbers_orig[i][comp] = 0;
					chars_orig[i][comp] = 99;
					if (number.equals("sin")==true) {
						chars_orig[i][comp] = 9;
						fun = true;
					}
					else if (number.equals("tan")==true) {
						chars_orig[i][comp] = 11;
						fun = true;
					}
					else if (number.equals("cos")==true) {
						chars_orig[i][comp] = 10;
						fun = true;
					}
					else if (number.charAt(0)=='y') {
						chars_orig[i][comp] = 12;
						try {
							numbers_orig[i][comp] = new Double(number.substring(1,number.length())).doubleValue();
						}
						catch (Exception e) {
							numbers_orig[i][comp] = 0;
						}
						/*Controller for variables y?*/
						if ((numbers_orig[i][comp] < 0) || (numbers_orig[i][comp] > components-1)) {
							status = false;
							error_msg="Bad input";
						}
						if (fun == true) {
							error_msg="Wrong function definition";
							status = false;
						}
					}
					else if (number.equals("t")==true) {
						chars_orig[i][comp] = 13;
						if (fun == true) {
							error_msg="Wrong function definition";
							status = false;
						}
					}
					else if (number.equals("log")==true) {
						chars_orig[i][comp] = 14;
						fun = true;
					}
					else if (number.equals("exp")==true) {
						chars_orig[i][comp] = 15;
						fun = true;
					}
					else if (number.equals("abs")==true) {
						chars_orig[i][comp] = 16;
						fun = true;
					}
					else if (number.equals("sqrt")==true) {
						chars_orig[i][comp] = 17;
						fun = true;
					}
					else if (number.equals("asin")==true) {
						chars_orig[i][comp] = 18;
						fun = true;
					}
					else if (number.equals("acos")==true) {
						chars_orig[i][comp] = 19;
						fun = true;
					}
					else if (number.equals("atan")==true) {
						chars_orig[i][comp] = 20;
						fun = true;
					}
					else if (number.equals("sign")==true) {
						chars_orig[i][comp] = 21;
						fun = true;
					}
					else {
						try {
							numbers_orig[i][comp]=new Double(number).doubleValue();
							chars_orig[i][comp]=99;
							if (fun == true) {
								error_msg="Wrong function definition";
								status = false;
							}
						}
						catch (Exception e) {
							status = false;
							error_msg="Bad input";
						}
					}
					i++;					
				}
				number=expression.substring(oper_i,oper_i+1).trim();
			   	if (number.equals("^")==true) {
					chars_orig[i][comp] = 1;
					if (fun == true) {
						error_msg="Wrong function definition";
						status = false;
					}
				}
				else if (number.equals("*")==true) {
					chars_orig[i][comp] = 2;
					if (fun == true) {
						error_msg="Wrong function definition";
						status = false;
					}
				}
				else if (number.equals("/")==true) {
					chars_orig[i][comp] = 3;
					if (fun == true) {
						error_msg="Wrong function definition";
						status = false;
					}
				}
				else if (number.equals("+")==true) {
					chars_orig[i][comp] = 4;
					if (fun == true) {
						error_msg="Wrong function definition";
						status = false;
					}
				}
				else if (number.equals("-")==true) {
					chars_orig[i][comp] = 5;
					if (fun == true) {
						error_msg="Wrong function definition";
						status = false;
					}
				}
				else if (number.equals("=")==true) {
					chars_orig[i][comp] = 6;
					if (fun == true) {
						error_msg="Wrong function definition";
						status = false;
					}
				}
				else if (number.equals("(")==true) {
					chars_orig[i][comp] = 7;
					fun = false;
				}
				else if (number.equals(")")==true)
					chars_orig[i][comp] = 8;
				numbers_orig[i][comp]=0;
				if ((chars_orig[i][comp] <7 ) && (chars_orig[i-1][comp] < 7)) {
						error_msg="Bad input";
						status = false;
				}
				i++;
				oper_i++;
				begin_num = oper_i;				
				if (number.equals("("))
					left_brac++;
				else if (number.equals(")")) {
					if (chars_orig[i-2][comp] < 7) {
						error_msg="Bad input";
						status = false;
					}
					right_brac++;
				}
			}
		}
		if (left_brac != right_brac && status==true) {
			error_msg="Unbalanced brackets";
			status = false;
		}
		return status;
	}
/*Starts the evaluation. Uses evalComp to evaluate single expressions*/	
	public final double[] evaluate(double t, double[] y) {
		error=false;
		int i,j,k;
		double[] solution = new double[y.length];
		for (j = 0 ; j < exp_orig.length()+5; j++) {
			for (k = 0; k < components; k++) {
				chars[j][k] = chars_orig[j][k];
				numbers[j][k] = numbers_orig[j][k];			
			}
		}
		for (i = 0 ; i < components ; i++)
			solution[i] = evalComp(t,y,i);
		return solution;
	}

/*Finds the innermost brackets and fnctions and gives these to*/
/*evalBrac and calc to be evaluated*/

	public final double evalComp(double x, double[] y, int comp) {
		int trigind = -2;
		int i = 0;
		int first_brac = 0;
		int second_brac = 0;
		while (chars[i][comp] != 6) {
			switch (chars[i][comp]) {
				case 13:
					chars[i][comp] = 99;
					numbers[i][comp] = x;
					break;
				case 12:
					chars[i][comp] = 99;
					numbers[i][comp] = y[(int)(numbers[i][comp])];
					break;
				case 7:
					first_brac = i;
					break;
				case 9:
					trigind = i;
					break;
				case 10:
					trigind = i;
					break;
				case 11:
					trigind = i;
					break;
				case 14:
					trigind = i;
					break;
				case 15:
					trigind = i;
					break;
				case 16:
					trigind = i;
					break;
				case 17:
					trigind = i;
					break;
				case 18:
					trigind = i;
					break;
				case 19:
					trigind = i;
					break;
				case 20:
					trigind = i;
					break;
				case 21:
					trigind = i;
					break;
				case 8:
					second_brac = i;
					chars[first_brac][comp] = 99;
					numbers[first_brac][comp] = evalBrac(first_brac + 1, second_brac - 1, comp);
					i = second_brac;
					if (trigind == first_brac - 1)
					{
						numbers[trigind][comp]= calc(numbers[first_brac][comp],1,chars[trigind][comp]);
						first_brac--;
					}
					while (chars[i][comp] != 6)
					{
						i++;
						int indtemp = i - second_brac + first_brac;
						numbers[indtemp][comp] = numbers[i][comp];
						chars[indtemp][comp] = chars[i][comp];
					}
					i = -1;
					break;
			}
			i++;
		}
		return numbers[0][comp];
	}

/*Handles the order ^, * /,+ -*/
	public final double evalBrac(int begin, int end, int comp) {
		if (chars[begin][comp]==5) {
			begin++;
			numbers[begin][comp] = 0 - numbers[begin][comp];
		}
		evalBasic(1,1, begin, end, comp);
		evalBasic(2,3, begin, end, comp);
		return evalBasic(4,5, begin, end, comp);
	}
	
/*Calculates from left to right given operations from given expression*/
	public final double evalBasic(int oper_begin, int oper_end, int begin, int end, int comp) {
	 	int i;
		int oper = begin;
		while (oper <= end) {
			if ((chars[oper][comp] >= oper_begin) && (chars[oper][comp] <= oper_end)){
				numbers[oper-1][comp] = calc(numbers[oper-1][comp],numbers[oper+1][comp],chars[oper][comp]);
				i = oper+1;
				while (i <= end)
				{
					i++;
					numbers[i-2][comp] = numbers[i][comp];
					chars[i-2][comp] = chars[i][comp];
				}
				oper = begin;
			}	
			oper++;
		}
		return numbers[begin][comp];
	}

/*Calculates the result of two numbers with given operator. In case of functions*/
/*only one number is used.*/
	public final double calc(double number1, double number2, int oper)
	{
		double result=0;
		switch (oper) {
			case 2:
				result = (number1*number2);
				break;
			case 3:
				result=(number1/number2);
				break;
			case 4:
				result = (number1+number2);
				break;
			case 5:
				result = (number1-number2);
				break;
			case 1:
				result=Math.pow(number1,number2);
				break;
			case 9:
				result = Math.sin(number1);
				break;
			case 10:
				result = Math.cos(number1);
				break;
			case 11:
				result = Math.tan(number1);
				break;
			case 14:
				result = Math.log(number1);
				break;
			case 15:
				result = Math.exp(number1);
				break;
			case 16:
				result = Math.abs(number1);
				break;
			case 17:
				result = Math.sqrt(number1);
				break;
			case 18:
				result = Math.asin(number1);
				break;
			case 19:
				result = Math.acos(number1);
				break;
			case 20:
				result = Math.atan(number1);
				break;
			case 21:
				if (number1 < 0) result = -1;
				else if (number1 > 0) result = 1;
				else result = 0;
				break;
		}
		if (Double.isInfinite(result) || Double.isNaN(result)) {
			error=true;
			error_msg="Inf or NaN";
		}					
		return result;
	}
}